/*
 *Andrew
 *Editing of flash GOTStateMachine example library
 *for potential use as our complex square / object avoidance base code
 *Trying to use State Machine coding to have cleaner / better code that is easier to read / follow.
 *Seems to be a good idea for robots. Examples of StateMachine programming are washing machines and stop lights. Defined states.
 *
 *4-28-2020 making progress. IR seems to now work.
 *I am now reffering to this code base as a "hybrid state machine".
 *As of now only the two states are defined as 'IdleState' and 'DriveInSquare'.
 *The rest are written as functions that can individually be called.
 *This method keeps the benefits of cleaner code writing of a state machine.
 *
 *still need to get global variable working for the forward encoders 
 *so that encoder compensation can occur when entering into the simple avoid routine.
 *As of now weird behavior happens when i try.
 *
 */

#include <GOTStateMachine.h> //include the GOTStateMachine library from github
//https://github.com/Grumpy-Old-Tech/GOTStateMachine

//add libraries needed for zumo
#include <Zumo32U4.h>
Zumo32U4Encoders encoders;
Zumo32U4Motors motors;
Zumo32U4LCD lcd;
Zumo32U4ButtonA buttonA;
Zumo32U4ProximitySensors proxSensors;

//define global variables for driving
#define FEET 2 //number of feet you want to travel in "forward" driving
#define mm (FEET / 0.0032808) //convert to mm
#define encoderticks (mm / 0.13) //convert to encoder ticks value
#define motorSpeed 250 //(up to about 400)
#define DISTANCE encoderticks
#define Kp 1
#define STRAIGHTFACTOR 1  // Adjust this to correct for minor curve.  Should be in the 0.9 to 1.1 range

volatile int IR = 0; //setting up simple variable for detecting if IR triggered or not

//setting encoders as global variables for the avoid code compensation
//currently not working correctly (screws up driving routine unexpectedly)
//volatile long DrivecountsLeft = encoders.getCountsAndResetLeft();
//volatile long DrivecountsRight = encoders.getCountsAndResetRight();


//***** State Machine *****
GOTStateMachine stateMachine(50); // execute every 50 milliseconds

// States
void IdleState();
void DriveInSquare();


void setup() {
  
  // Get the proximity sensors initialized
  proxSensors.initThreeSensors();
  pinMode(LED_BUILTIN, OUTPUT);

 // attachInterrupt(IR, SimpleAvoid, RISING); //tried to attach IR as an "external" intterupt, not sure if this is actually doing anything

  //Serial.begin(9600); //for debug only
  
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  stateMachine.setStartState(IdleState); // Initialise state machine
}

void loop() {

  stateMachine.execute();             // process the states as required
}

//***** States *****

void IdleState() {
  IR = 0; //reset IR flag to 0
  lcd.clear();
  lcd.gotoXY(0, 0);
  lcd.print("Press A");
  motors.setSpeeds(0, 0);
  buttonA.waitForButton();
  lcd.clear();
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  // Transitions
    stateMachine.changeState(DriveInSquare);
    return;
}

void DriveInSquare() {

//Drive in a Square
  lcd.clear();
  lcd.gotoXY(0, 0);
  lcd.print("Square");
  
  //run drive sequence here
  //drive in a 2ft x 2ft square
    forward(DISTANCE); //go forward # of defined FEET (set above)
    turnLeft(90); //turn left 90 degrees
    forward(DISTANCE);
    turnLeft(90);
    forward(DISTANCE);
    turnLeft(90);
    forward(DISTANCE);

  // Transitions
    stateMachine.changeState(IdleState);
    return;
}

//***** Functions ***** 

void IRcode(){
  
  //IR Proximity code here
// --- Read the sensors
  proxSensors.read();    
  int left_sensor = proxSensors.countsLeftWithLeftLeds();
  int center_left_sensor = proxSensors.countsFrontWithLeftLeds();
  int center_right_sensor = proxSensors.countsFrontWithRightLeds();
  int right_sensor = proxSensors.countsRightWithRightLeds();

  // Update the display
  lcd.gotoXY(0, 1);
  lcd.print(left_sensor);
  lcd.print(" ");
  lcd.print(center_left_sensor);
  lcd.print(" ");
  lcd.print(center_right_sensor);
  lcd.print(" ");
  lcd.print(right_sensor);

//Code for using multiple IR sensors in tandem to trigger when an obstacle is detected but not too sensitive and not too late.
//Uses the left and the front left sensors together
//Uses the right and front right sensors together
//When the left sensor detects a value above 5 AND the front-left sensor detects a value above 5 then an obstacle is detected
  if (left_sensor >= 5 && center_left_sensor >= 5){
    digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
    IR = 1; // in addition to the LED, set a flag when an obstacle is detected
  }
  else if (center_right_sensor >= 5 && right_sensor >= 5){
    digitalWrite(LED_BUILTIN, HIGH);   // turn the LED on (HIGH is the voltage level)
    IR = 1; // in addition to the LED, set a flag when an obstacle is detected
  }
  else {
    digitalWrite(LED_BUILTIN, LOW);    // turn the LED off by making the voltage LOW
    IR = 0; // no obstacles detected, reset flag to zero
  }

//Serial.println(IR); // used for debugging only. disable in final code when no longer needed.
}


// Go forward
void forward(long count) {
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  lcd.clear();
  lcd.gotoXY(0, 0);
  lcd.print("forward");

  //forward code
  int currentSpeedLeft=motorSpeed;
  int currentSpeedRight=motorSpeed;
  int error;
  int correction;

// attempting to make these global variables
// renamed countsLeft and countsRight to DrivecountsLeft / Right to prevent conflict with turn left / right routines.   
//  int countsLeft = encoders.getCountsAndResetLeft(); //original local variable
//  int countsRight = encoders.getCountsAndResetRight(); // original local variable
volatile long DrivecountsLeft = encoders.getCountsAndResetLeft(); //local variable
volatile long DrivecountsRight = encoders.getCountsAndResetRight(); //local variable


  motors.setSpeeds(currentSpeedLeft,currentSpeedRight);

//Obstacle Detection
 IRcode(); //run the IR proximity function here and then continue
//if IR flag is set
//turn off motors & run simple avoid function (and return after)
  if (IR == 1) {
      motors.setSpeeds(0,0);
      //IR = 0;
    SimpleAvoid();
  }
  
else{
while(DrivecountsLeft < count && DrivecountsRight < count)
{  
  
//Obstacle Detection
 IRcode(); //run the IR proximity function here and then continue
 //if IR flag is set
//turn off motors & run simple avoid function (and return after)
    if (IR == 1) {
      motors.setSpeeds(0,0);
      //IR = 0;
    SimpleAvoid();
  }
  
  else {
    DrivecountsLeft = encoders.getCountsLeft(); //get encoder value (do not reset)
    DrivecountsRight = encoders.getCountsRight(); //get encoder value (do not reset)
    error = DrivecountsLeft-STRAIGHTFACTOR*DrivecountsRight;
    correction = Kp*error;
    currentSpeedRight = motorSpeed + correction;
    motors.setSpeeds(currentSpeedLeft,currentSpeedRight);
  } 
}
  motors.setSpeeds(0,0);
encoders.getCountsAndResetLeft(); //reset encoder value
encoders.getCountsAndResetRight(); //reset encoder value

}
}

void reverse(long count) {
  lcd.clear();
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  long countsLeft = 0;
  long countsRight = 0;
  motors.setSpeeds(-motorSpeed, -motorSpeed);
  while(countsLeft < count) {
    countsLeft -= encoders.getCountsAndResetLeft();
    lcd.gotoXY(0, 0);
    lcd.print(countsLeft);
    lcd.gotoXY(0, 1);
    lcd.print(countsRight);
  }
    while(countsRight < count) {
    countsRight -= encoders.getCountsAndResetRight();
    lcd.gotoXY(0, 0);
    lcd.print(countsLeft);
    lcd.gotoXY(0, 1);
    lcd.print(countsRight);
  };
  
  motors.setSpeeds(0, 0);
}

// Turn left
void turnLeft(long angle) {
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  lcd.clear();
  lcd.gotoXY(0, 0);
  lcd.print("LEFT");

  //turn left using only the encoders!

  // Reset everything to zero
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  // Variables to keep track of stuff
  long countsLeft = 0;
  long countsRight = 0;
  //average of 725 Encoder Ticks per 90 degrees
  //this number may need to be dialed in
  //725 / 90 = 8.05
  //1 degree = 8.05 Encoder Ticks
int EC = (angle * 8.05);
  
  // Loop until we get there
  while(countsRight < EC) {

    // --- Read sensor values
    // Wheel encoders for distance
    countsLeft += encoders.getCountsAndResetLeft();
    countsRight += encoders.getCountsAndResetRight();
    
    int distanceLeft = EC - countsRight;

    // Set the motor speed.
    motors.setSpeeds(-motorSpeed, motorSpeed);

  }
}


// Turn right
void turnRight(long angle) {

  // Reset everything to zero
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  // Variables to keep track of stuff
  long countsLeft = 0;
  long countsRight = 0;
  //average of 725 Encoder Ticks per 90 degrees
  //this number may need to be dialed in
  //725 / 90 = 8.05
  //1 degree = 8.05 Encoder Ticks
int EC = (angle * 8.05);
  
  // Loop until we get there
  while(countsLeft < EC) {

    // --- Read sensor values
    // Wheel encoders for distance
    countsLeft += encoders.getCountsAndResetLeft();
    countsRight += encoders.getCountsAndResetRight();
    
    int distanceLeft = EC - countsLeft;

    // Set the motor speed.
    motors.setSpeeds(motorSpeed, -motorSpeed);

  }
}

void forwardAvoid(long count) {
  lcd.clear();
  encoders.getCountsAndResetLeft();
  encoders.getCountsAndResetRight();

  long countsLeft = 0;
  long countsRight = 0;
  motors.setSpeeds(motorSpeed, motorSpeed);
  while(countsLeft < count) {
    countsLeft += encoders.getCountsAndResetLeft();
    lcd.gotoXY(0, 0);
    lcd.print(countsLeft);
    lcd.gotoXY(0, 1);
    lcd.print(countsRight);
  }
    while(countsRight < count) {
    countsRight += encoders.getCountsAndResetRight();
    lcd.gotoXY(0, 0);
    lcd.print(countsLeft);
    lcd.gotoXY(0, 1);
    lcd.print(countsRight);
  };
  
  motors.setSpeeds(0, 0);
}

void SimpleAvoid() {
  lcd.clear();
  lcd.gotoXY(0, 0);
  lcd.print("Avoid");

//  //Encoder Avoid Code Compensation begin
//  int OldEncodersLeft;
//  int OldEncodersRight;
//  OldEncodersLeft = DrivecountsLeft;
//  OldEncodersRight = DrivecountsRight;

  // Simple Avoid Main Routine
    turnRight(90); //turn left 90 degrees
    forwardAvoid(1000);
    turnLeft(90); //turn left 90 degrees
    forwardAvoid(2000);
    turnLeft(90); //turn left 90 degrees
    forwardAvoid(1000);
    turnRight(90); //turn left 90 degrees
    motors.setSpeeds(0, 0);

//  //Encoder Avoid Code Compensation continue
//  int NewEncodersLeft;
//  int NewEncodersRight;
//  NewEncodersLeft = encoders.getCountsLeft();
//  NewEncodersRight = encoders.getCountsRight();
//  //do math to reset encoders to value they were before entering the avoid routine
//  int EncodersCompensateLeft;
//  int EncodersCompensateRight;
//  EncodersCompensateLeft = NewEncodersLeft - (NewEncodersLeft - OldEncodersLeft);
//  DrivecountsLeft = EncodersCompensateLeft;
//  EncodersCompensateRight = NewEncodersRight - (NewEncodersRight - OldEncodersRight);
//  DrivecountsRight = EncodersCompensateRight;

     IR = 0; //reset IR flag to 0
}
